@@ -1,6 +1,5 @@
 from dataclasses import dataclass, replace
 import numpy as np
-from scipy.spatial.distance import cdist
 from scipy.special import expit
 import os
 os.environ['TF_CPP_MIN_LOG_LEVEL'] = '1'
@@ -13,6 +12,7 @@ with warnings.catch_warnings():
     import tensorflow_probability as tfp
 
 from .spatialinterpolator import SpatialInterpolator
+from .covfunc import CovarianceFunction
 
 MVN = tfp.distributions.MultivariateNormalTriL
 
@@ -45,62 +45,39 @@ class NormalizingFeaturizer:
         F_norm = (F_unnorm - self.unnorm_mean) / self.unnorm_std
         return np.concatenate([ones, F_norm], axis=1)
 
+# Other needed functions.
+def logodds(p):
+    return -np.log(1/p-1)
+
+def logodds_half(x):
+    return logodds(x/2)
+
 def e(x, a=-1):
     return tf.expand_dims(x, a)
 
-def one_axes(x):
-    return tf.where(tf.equal(x.shape, 1))[:, 0]
-
-# Squared exponential covariance function.
-def gp_covariance_sq_exp(D, F, range_, sill, nugget, alpha):
-    C = sill * tf.exp(-tf.square(D / range_)) \
-            + (nugget + 1e-6) * tf.eye(D.shape[0], dtype=tf.float64) \
-            + alpha * tf.einsum('ba,ca->bc', F, F)
-    return C
-
-# Gamma exponential covariance function.
-@tf.custom_gradient
-def safepow(x, a):
-    y = tf.pow(x, a)
-    def grad(dy):
-        dx = tf.where(x <= 0.0, tf.zeros_like(x), dy * tf.pow(x, a-1))
-        dx = tf.reduce_sum(dx, axis=one_axes(x), keepdims=True)
-        da = tf.where(x <= 0.0, tf.zeros_like(a), dy * y * tf.math.log(x))
-        da = tf.reduce_sum(da, axis=one_axes(a), keepdims=True)
-        return dx, da
-    return y, grad
-
-def gamma_exp(d2, halfgamma):
-    return tf.exp(-safepow(tf.maximum(d2, 0.0), halfgamma))
-
-def gp_covariance_gamma_exp(D, F, range_, sill, nugget, halfgamma, alpha):
-    range_ = e(e(range_))
-    sill = e(e(sill))
-    halfgamma = e(e(halfgamma))
-    C = sill * gamma_exp(tf.square(D / range_), halfgamma) + (nugget + 1e-6) * tf.eye(tf.shape(D)[0], dtype=tf.float64)
-    C += tf.einsum('ba,ca->bc', F, F) * alpha
-    return C
+def gp_covariance(covariance, X, F, p, alpha):
+    X = tf.cast(X, tf.float32)
+    F = tf.cast(F, tf.float32)
+    C = covariance.matrix(X, p)
+    C += 1e-6 * tf.eye(X.shape[0])
+    C += alpha * tf.einsum('ba,ca->bc', F, F)
+    return tf.cast(C, tf.float64)
 
 # Transform parameters.
 @tf.function
-def gp_xform_parameters(parameters, covariance_func):
+def gp_xform_parameters(up):
     """
     Transform parameters from the underlying representation
     (which has the whole real number line as its range) to
     a surface representation (which is bounded).
     """
-    range_ = tf.exp(parameters['log_range'])
-    sill = tf.exp(parameters['log_sill'])
-    nugget = tf.exp(parameters['log_nugget'])
-
-    if covariance_func == 'squared-exp':
-        param_dict = {'range': range_, 'sill': sill, 'nugget': nugget}
-        return param_dict
-
-    elif covariance_func == 'gamma-exp':
-        halfgamma = tf.sigmoid(parameters['logit_halfgamma'])
-        param_dict = {'range': range_, 'sill': sill, 'nugget': nugget, 'halfgamma': halfgamma}
-        return param_dict
+    sp = {}
+    for v in up.keys():
+        if v == 'gamma':
+            sp[v] = 2.0 * tf.sigmoid(up[v])
+        else:
+            sp[v] = tf.exp(up[v])
+    return sp
 
 # Log likelihood.
 @tf.function
@@ -109,27 +86,25 @@ def gp_log_likelihood(u, m, cov):
     u_adj = u - m
     logdet = tf.linalg.logdet(2 * np.pi * cov)
     quad = tf.matmul(e(u_adj, 0), tf.linalg.solve(cov, e(u_adj, -1)))[0, 0]
-    return -0.5 * (logdet + quad)
+    return tf.cast(-0.5 * (logdet + quad), tf.float32)
 
 
 # GP training.
-def gpm_train_step(optimizer, data, parameters, hyperparameters, covariance_func):
+def gpm_train_step(optimizer, data, parameters, hyperparameters, covariance):
     with tf.GradientTape() as tape:
-        p = gp_xform_parameters(parameters, covariance_func)
+        p = gp_xform_parameters(parameters)
         beta_prior = hyperparameters['alpha']
 
-        if covariance_func == 'squared-exp':
-            A = gp_covariance_sq_exp(data['D'], data['F'], p['range'], p['sill'], p['nugget'], beta_prior)
-        elif covariance_func == 'gamma-exp':
-            A = gp_covariance_gamma_exp(data['D'], data['F'], p['range'], p['sill'], p['nugget'], p['halfgamma'], beta_prior)
+        A = gp_covariance(covariance, data['X'], data['F'], p, hyperparameters['alpha'])
 
         ll = gp_log_likelihood(data['u'], 0., A)
 
         if hyperparameters['reg'] != None:
-            reg = hyperparameters['reg'] * tf.reduce_sum(tf.square(parameters['log_range']))
-            loss = -ll + reg
+            reg = hyperparameters['reg'] * covariance.reg(p)
         else:
-            loss = -ll
+            reg = 0.
+
+        loss = -ll + reg
 
     gradients = tape.gradient(loss, parameters.values())
     optimizer.apply_gradients(zip(gradients, parameters.values()))
@@ -139,7 +114,7 @@ def gpm_train_step(optimizer, data, parameters, hyperparameters, covariance_func
 class GP(SpatialInterpolator):
 
     featurizer: NormalizingFeaturizer
-    covariance_func: str = 'squared-exp'
+    covariance: CovarianceFunction
     parameters: object = None
     hyperparameters: object = None
     verbose: bool = True
@@ -159,7 +134,7 @@ class GP(SpatialInterpolator):
                                 return x1[:, 0], x1[:, 1], x1[:, 0]**2, x1[:, 1]**2.
                     Default is None.
 
-                covariance_func : str
+                covariance : CovarianceFunction
                      Name of the covariance function to use in the GP.
                      Should be 'squared-exp' or 'gamma-exp'.
                      Default is 'squared-exp'.
@@ -188,48 +163,27 @@ class GP(SpatialInterpolator):
         default_hyperparameters = dict(alpha=10, reg=None, train_iters=300)
         self.hyperparameters = dict(default_hyperparameters, **self.hyperparameters)
 
-        # Default parameters.
-        if self.parameters == None:
-            if self.covariance_func == 'squared-exp':
-                self.parameters = dict(range=1.0, sill=1.0, nugget=1.0)
-            elif self.covariance_func == 'gamma-exp':
-                self.parameters = dict(range=1.0, sill=1.0, nugget=1.0, gamma=1.0)
-
-        # Other needed functions.
-        def logodds(p):
-            return -np.log(1/p-1)
-
-        def logodds_half(x):
-            return logodds(x/2)
-
         # def sigmoid(o): return 1 / (1 + np.exp(-o))
         # def doub_sigmoid(o): return 2 * sigmoid(o)
 
         # List the needed graph functions.
 
-        # Set the user desired covariance function.
-        if self.covariance_func not in ['squared-exp', 'gamma-exp']:
-            raise ValueError("Only 'squared-exp' and 'gamma-exp' are currently supported.")
-
         # Define other inputs.
         self.train_iters = self.hyperparameters['train_iters']
 
         if self.hyperparameters['reg']:
-            self.hp = {'alpha': tf.constant(self.hyperparameters['alpha'], dtype=tf.float64),
-                                    'reg': tf.constant(self.hyperparameters['reg'], dtype=tf.float64)}
+            self.hp = {'alpha': tf.constant(self.hyperparameters['alpha'], dtype=tf.float32),
+                       'reg': tf.constant(self.hyperparameters['reg'], dtype=tf.float32)}
         if not self.hyperparameters['reg']:
-            self.hp = {'alpha': tf.constant(self.hyperparameters['alpha'], dtype=tf.float64),
-                                    'reg': None}
+            self.hp = {'alpha': tf.constant(self.hyperparameters['alpha'], dtype=tf.float32),
+                       'reg': None}
 
     def fit(self, x, u):
-        # Distance matrix.
-        D = cdist(x, x)
-
         # Feature matrix.
         F = self.featurizer(x)
 
         # Data dict.
-        self.data = {'D': tf.constant(D), 'F': tf.constant(F), 'u': tf.constant(u)}
+        self.data = {'X': tf.constant(x), 'F': tf.constant(F), 'u': tf.constant(u)}
 
         # Train the GP.
         def gpm_fit(data, parameters, hyperparameters):
@@ -238,17 +192,13 @@ class GP(SpatialInterpolator):
             j = 0 # Iteration count.
             for i in range(10):
                 while j < (i + 1) * self.train_iters / 10:
-                    p, ll = gpm_train_step(optimizer, data, parameters, hyperparameters, self.covariance_func)
+                    p, ll = gpm_train_step(optimizer, data, parameters, hyperparameters, self.covariance)
                     j += 1
 
                 if self.verbose == True:
-                    if self.covariance_func == 'squared-exp':
-                        print('[iter %d] [ll %7.2f] [range %4.2f, sill %4.2f, nugget %4.2f]' %
-                            (j, ll, p['range'], p['sill'], p['nugget']))
-
-                    elif self.covariance_func == 'gamma-exp':
-                        print('[iter %d] [ll %7.2f] [range %4.2f, sill %4.2f, nugget %4.2f, gamma %4.2f]' %
-                            (j, ll, p['range'], p['sill'], p['nugget'], p['halfgamma'] * 2))
+                    s = '[iter %4d, ll %7.2f]' % (j, ll)
+                    s += self.covariance.report(p)
+                    print(s)
 
         up = self.get_underlying_parameters()
 
@@ -257,43 +207,33 @@ class GP(SpatialInterpolator):
         return replace(self, parameters = self.get_surface_parameters(up))
 
     def get_underlying_parameters(self):
-        if self.covariance_func == 'squared-exp':
-            up = dict(
-                log_range = tf.Variable(np.log(self.parameters['range']), dtype=tf.float64),
-                log_sill = tf.Variable(np.log(self.parameters['sill']), dtype=tf.float64),
-                log_nugget = tf.Variable(np.log(self.parameters['nugget']), dtype=tf.float64))
-        elif self.covariance_func == 'gamma-exp':
-            up = dict(
-                log_range = tf.Variable(np.log(self.parameters['range']), dtype=tf.float64),
-                log_sill = tf.Variable(np.log(self.parameters['sill']), dtype=tf.float64),
-                log_nugget = tf.Variable(np.log(self.parameters['nugget']), dtype=tf.float64),
-                logit_halfgamma = tf.Variable(logodds_half(self.parameters['gamma']), dtype=tf.float64))
+        up = {}
+        for v in self.parameters.keys():
+            if v == 'gamma':
+                up[v] = tf.Variable(logodds_half(self.parameters[v]), dtype=tf.float32)
+            else:
+                up[v] = tf.Variable(np.log(self.parameters[v]), dtype=tf.float32)
         return up
 
     def get_surface_parameters(self, up):
         sp = {}
-        sp['range'] = np.exp(up['log_range'].numpy())
-        sp['sill'] = np.exp(up['log_sill'].numpy())
-        sp['nugget'] = np.exp(up['log_nugget'].numpy())
-        if 'logit_halfgamma' in up:
-            sp['gamma'] = 2.0 * expit(parameters['logit_halfgamma'])
+        for v in up.keys():
+            if v == 'gamma':
+                sp[v] = 2.0 * expit(up[v])
+            else:
+                sp[v] = np.exp(up[v])
         return sp
 
     def generate(self, x):
-        xr = x.reshape([-1, x.shape[-1]])
+        X = x.reshape([-1, x.shape[-1]])
 
-        D = cdist(xr, xr)
-
-        F = self.featurizer(xr)
+        F = self.featurizer(X)
 
         up = self.get_underlying_parameters()
 
-        p = gp_xform_parameters(up, self.covariance_func)
+        p = gp_xform_parameters(up)
 
-        if self.covariance_func == 'squared-exp':
-            A = gp_covariance_sq_exp(D, F, p['range'], p['sill'], p['nugget'], self.hyperparameters['alpha'])
-        elif self.covariance_func == 'gamma-exp':
-            A = gp_covariance_gamma_exp(D, F, p['range'], p['sill'], p['nugget'], p['halfgamma'], self.hyperparameters['alpha'])
+        A = gp_covariance(self.covariance, X, F, p, self.hyperparameters['alpha'])
 
         z = tf.zeros_like(A[0, :])
 
@@ -327,27 +267,18 @@ class GP(SpatialInterpolator):
         def e(x, a=-1):
             return tf.expand_dims(x, a)
 
-        def one_axes(x):
-            return tf.where(tf.equal(x.shape, 1))[:, 0]
-
-        ##############################################
         def interpolate_gp(X1, u1, X2, parameters, hyperparameters):
 
             N1 = len(X1) # Number of measurements.
             N2 = len(X2) # Number of predictions.
 
             X = np.concatenate([X1, X2], axis=0)
-            D = cdist(X, X)
 
-            # Use the user given featurization.
             F = self.featurizer(X)
 
-            p = gp_xform_parameters(parameters, self.covariance_func)
+            p = gp_xform_parameters(parameters)
 
-            if self.covariance_func == 'squared-exp':
-                A = gp_covariance_sq_exp(D, F, p['range'], p['sill'], p['nugget'], hyperparameters['alpha'])
-            elif self.covariance_func == 'gamma-exp':
-                A = gp_covariance_gamma_exp(D, F, p['range'], p['sill'], p['nugget'], p['halfgamma'], hyperparameters['alpha'])
+            A = gp_covariance(self.covariance, X, F, p, hyperparameters['alpha'])
 
             A11 = A[:N1, :N1]
             A12 = A[:N1, N1:]
